@@ -385,7 +385,7 @@ pub trait Node: Any {
   //}
   /// Return the children of the node. If the node has no children then returns an empty vector.
   fn children(&self) -> Vec<Rc<dyn Node>>;
-  /// Return descendants of the Node, but including the Node itself.
+  /// Return descendants of the Node, but not including the Node itself.
   fn descendants(&self) -> Vec<Rc<dyn Node>>;
   /// Return the next following sibling.
   fn get_following_sibling(&self) -> Option<Rc<dyn Node>>;
@@ -428,6 +428,9 @@ pub trait Node: Any {
   fn add_child(&self, c: &dyn Any) -> Result<(), Error>;
   /// Add a text node as a child.
   fn add_text_child(&self, t: String) -> Result<(), Error>;
+
+  /// Remove a node from its parent
+  fn remove(&self) -> Result<(), Error>;
 }
 
 #[derive(Clone, Debug)]
@@ -459,6 +462,15 @@ impl QualifiedName {
   pub fn get_localname(&self) -> String {
     self.localname.clone()
   }
+  pub fn to_string(&self) -> String {
+    let mut result = String::new();
+    self.prefix.as_ref().map_or((), |p| {
+      result.push_str(p.as_str());
+      result.push(':');
+    });
+    result.push_str(self.localname.as_str());
+    result
+  }
 }
 
 struct Ancestor {
@@ -1293,4 +1305,15 @@ mod tests {
     fn op_after() {
       assert_eq!(Operator::After.to_string(), ">>")
     }
+
+    // QualifiedName
+
+    #[test]
+    fn qn_unqualified() {
+      assert_eq!(QualifiedName::new(None, None, "foo".to_string()).to_string(), "foo")
+    }
+    #[test]
+    fn qn_qualified() {
+      assert_eq!(QualifiedName::new(Some("http://example.org/whatsinaname/".to_string()), Some("x".to_string()), "foo".to_string()).to_string(), "x:foo")
+    }
 }
